struct account {
	const char *name;
	int balance;
	struct lock lock;
};

int num_accounts;
struct account *accounts;

bool transfer(int amount, struct account *from, struct account *to) {
	lock_acquire(&from->lock);
	bool ok = from->balance >= amount;
	if (ok)
		from->balance -= amount;
	lock_release(&from->lock);
	lock_acquire(&to->lock);
	if (ok)
		to->balance += amount;
	lock_release(&to->lock);
	return ok;
}

void worker(int amount, int from, int to) NO_STEP {
	transfer(amount, &accounts[from], &accounts[to]);
}

int total_money() NO_STEP {
	int total = 0;
	for (int i = 0; i < num_accounts; i++)
		total += accounts[i].balance;
	return total;
}

int main(void) {
	NO_STEP {
		num_accounts = 4;
		accounts = malloc(sizeof(struct account) * num_accounts);
		accounts[0].balance = 20;
		accounts[1].balance = 25;
		accounts[2].balance = 15;
		for (int i = 0; i < num_accounts; i++)
			lock_init(&accounts[i].lock);
	}

	thread_new(&worker, 10, 1, 2);
	thread_new(&worker, 20, 1, 2);

	// What about this?
	worker(10, 0, 3);

	printf("Total amount of money: %d\n", total_money());

	return 0;
}
